#ifndef __PLUGIN_H__
#define __PLUGIN_H__

#include "plugin_type_info.h"
#include <utility>

namespace utility
{

class plugin;

// Creates multiple types of plug-ins.
class factory
{
public:
	typedef size_t size_type;

	// Returns the number of factory types available.
	virtual size_type size() const throw() = 0;
	// Creates a plug-in object of the given type (between 0 and length() - 1).  Returns NULL
	// if there is an error.  The object returned will be destroy()'ed when no longer needed.
	virtual plugin* create_plugin(size_type index) const throw() = 0;
	// Returns a pointer to the type-info of the given index (between 0 and length() - 1).
	// Returns NULL if there is an error.
	virtual const utility::plugin_type_info* plugin_typeid(size_type index) const throw() = 0;

	// Handle to a factory.  Simplifies the task of loading and unloading plugin libraries.
	// A factory handle *must* remain in existence as long as any plugins created by it exist.
	class handle
	{
	public:
		// Constructs invalid handle.
		handle() throw() : f(0), data(0) {}
		// Constructs handle with factory loaded from given file.
		explicit handle(const char* path) throw();
		// Copies the handle.  A library remains loaded until *all* its handles are destroyed.
		handle(const handle& c) throw() : f(0), data(0) { *this = c; }
		// Unloads the library if this is the last handle to the loaded library.
		~handle() throw();
		// Assigns this handle to another.  If this handle is valid, it is unload()ed prior to assignment.
		handle& operator=(const handle& c) throw();
		// Explicitly unloads the library.  Makes this handle invalid.
		void unload() throw();
		// Returns true if the handle is valid, false otherwise.
		bool valid() const throw()  { return f != NULL; }
		// Both allow access to the underlying factory object.  The factory handle must be valid.
		const factory* operator->() const throw()  { return f; }
		const factory& operator*() const throw()  { return *f; }
	protected:
		const factory* f;
		void* data; 
	};
};

// All plug-in objects must derive from this class.
class plugin
{
public:
	// All plugin derivatives must declare either a native or an alien interface.
	__PLUGIN_INTERFACE0(1, plugin);
	// Calls the objects destructor and deallocates its memory.  This function is called as opposed to
	// explicitly deleting (with delete) object, since allocation mechanisms usually cannot span dll-boundaries.
	virtual void destroy() throw() = 0;
	// Creates a copy of this plugin, or returns NULL if it cannot.  The plugin returned will be destroy()'ed
	// when no longer needed.
	virtual plugin* clone() const throw() = 0;
	// Creates a new instance of this plugin's type (not a copy).
	virtual plugin* instance() const throw() = 0;
};

// Function pointer to a "get_factory" function.  All external plug-in libraries must
// export a function of this type.  I.e. all external plug-in libraries must define the
// following function:
// __PLUGIN_EXPORT const factory* get_factory() { ... }
#define __PLUGIN_EXPORT extern "C" __declspec(dllexport)
static const char fp_get_factory_name[] = "get_factory";
typedef const factory* (*fp_get_factory)();

// Loads a factory from a plugin dll.  Returns an invalid handle if the plugin dll cannot be loaded.
// Alternately, the factory::handle constructor can be used directly.
inline factory::handle load_factory(const char* path)  { return factory::handle(path); }

}; // namespace utility

#endif  //#ifndef __PLUGIN_H__

